/* Utilities for Samtools.   Version 1.2, 22 March 1996 */
#include<fcntl.h>
#include<sys/types.h>
#include<unistd.h>
#include<stdlib.h>
#include<errno.h>
#include<ctype.h>
#define caseeq(x,y) ((x)==(y) || \
        (isascii(x)&&isascii(y)&&toupper(x)==toupper(y)))

#ifdef SUN_DKIO
#include<sun/dkio.h>
#endif

#include"sam.h"

unsigned char map[MAP_LEN];
dir_entry *directory;
unsigned char *matched;
struct _matchinfo matchinfo[MATCHES];
int diskfree;
char *progname;
char *device=0;
int Mdos=0;
int dirno=0;
int ign_dirno=0;
int dir_entries;
char slash;
char *getenv();

int samdir(mode)
char *mode;
{
   int s,i,j,m,len,rlen;
#ifdef SUN_DKIO
   struct fdk_char cha;
   struct dk_map dmap;
   struct stat stinfo;
#endif
   unsigned char *dptr;
   char *tmp;
   int flag=mode[0]=='w' ? O_RDWR : O_RDONLY;
   int f;
   directory=(dir_entry *)malloc(512);
   if(!directory){
      fprintf(stderr,"%s: cannot allocate enough memory\n",progname);
      exit(1);
   }
   if(!device)device=getenv(ENVVAR);
   if(!device)device=DEVICE;
   f=open(device,flag);
   if(f<0)samerror("Couldn't open",device);

#ifdef SUN_DKIO
   /* Configure the device for 10 sectors, 80 tracks, if it is a
      character-special device. */
   if(fstat(f,&stinfo))
      samerror("Couldn't stat fd after opening",device);
   if(S_ISCHR(stinfo.st_mode)){
      if(ioctl(f,FDKIOGCHAR,&cha))
         samerror("Error from get-characteristics ioctl on",device);
      if(ioctl(f,DKIOCGPART,&dmap))
         samerror("Error from get-partition ioctl on",device);
      cha.secptrack=SECTORS; cha.ncyl=NCYL;
      dmap.dkl_nblk=NBLK;
      if(ioctl(f,FDKIOSCHAR,&cha))
         samerror("Error from set-characteristics ioctl on",device);
      if(ioctl(f,DKIOCSPART,&dmap))
         samerror("Error from set-partition ioctl on",device);
   }
#endif
   /* In case we couldn't configure, do a cursory check to make sure
      the device contains 80 tracks of 10 sectors */
   if(lseek(f,(NBLK-1)*SECTOR_LEN,SEEK_SET)<0)
      samerror("lseek failed on",device);
   if(read(f,&directory[0][0],512)<512){
      fprintf(stderr,"%s: error: file %s too short, or not enough sectors on device\n",
         progname,device);
      exit(1);
   }
   if(read(f,&directory[0][0],512)>0)
      fprintf(stderr,"%s: warning: file %s too long, or too many sectors on device\n",
         progname,device);
   if(lseek(f,0,SEEK_SET)<0)
      samerror("lseek failed on",device);
   if(read(f,&directory[0][0],512)<0)
      samerror("read failed on",device);
   Mdos = directory[0][210]>31 && directory[0][255]<40 &&
          (directory[0][252] || directory[0][253]);
   if(Mdos)dir_entries=DIR_PER_TRACK*(DIR_TRACKS+directory[0][255]);
   else dir_entries=DIR_PER_TRACK*DIR_TRACKS;
   free(directory);
   matched=(unsigned char *)malloc(dir_entries);
   directory=(dir_entry *)malloc(dir_entries*DIR_LEN);
   if(!matched || !directory){
      fprintf(stderr,"%s: cannot allocate enough memory\n",progname);
      exit(1);
   }
   fputs("Reading directory...",stdout);fflush(stdout);
   dptr=&directory[0][0];
   for(s=0;s<dir_entries/2;s+=10){
      findsector(f,s);
      len=DIR_PER_TRACK*DIR_LEN;
      while(len){
         if((rlen=read(f,dptr,len))<0)samerror("Couldn't read",device);
         else if(!rlen)errno=EOF_MAGIC,samerror("Couldn't read",device);
         else len-=rlen,dptr+=rlen;
      }
   }
   puts("done.");
   diskfree=FREE;
   for(s=0;s<dir_entries;s++){
      if(s==RESERVED)s+=2;
      if(directory[s][0]){
         diskfree-=directory[s][11]*256+directory[s][12];
         for(i=0;i<MAP_LEN;i++)map[i]|=directory[s][i+15];
      }
   }
   dirno=0;
   tmp=getenv(SLASHVAR);
   if(tmp && tmp[0])slash=tmp[0];
   else slash='/';
   if(Mdos){
      i=0;m=1;
      for(s=0;s<directory[0][255];s++)for(j=0;j<10;j++){
         if((s||j) && !(map[i]&m)){
            diskfree--;
            map[i]|=m;
         }
         if((m<<=1)==256)m=1,i++;
      }
      tmp=getenv(DIRVAR);
      if(tmp && tmp[0])dirno=finddir(tmp,1);
   }
   return f;
}

int match(pattern,name)
char *pattern, *name;
{
   int i,j,p,newi,newj;
   char c;
   i=j=p=0;
   while((c=pattern[i++])){
      if(j==NAME_LEN)return 1; /* a too-long pattern matches, for now. */
      if(c=='?'){
         matchinfo[p].t='?';
         matchinfo[p++].x=j++;
      }
      else if(c!='*')
         if(caseeq(c,name[j]))j++;
         else return 0;
      else{
         matchinfo[p].t='*';
         matchinfo[p].x=j;
         matchinfo[p].y=j;
         while(j<NAME_LEN){
            newj=j; newi=i;
            while((c=pattern[newi++]) && c!='*'){
               if(newj==NAME_LEN)return 0; /* too-long wildcard */
               if(c=='?')newj++;
               else if(caseeq(c,name[newj]))newj++;
               else break;
            }
            if(!c){
               while(newj<NAME_LEN && name[newj]==' ')newj++;
               if(newj==NAME_LEN)break;
            }
            if(c=='*')break;
            matchinfo[p].y=++j;
         }
         if(j>NAME_LEN)return 0; /* a too-long wildcard fails, for now */
         if(j==NAME_LEN && c && c!='*')return 0;
         p++;
      }
   }
   while(j<NAME_LEN)if(name[j++]!=' ')return 0;
#if 0
   if(p)printf("Matched %s with %10.10s:\n",pattern,name);
   for(i=0;i<p;i++){
      printf("'%c' matched '",matchinfo[i].t);
      if(matchinfo[i].t=='?')putchar(name[matchinfo[i].x]);
      else for(j=matchinfo[i].x;j<matchinfo[i].y;j++)
         putchar(name[j]);
      putchar('\'');
      putchar('\n');
   }
#endif
   return 1;
}

int matchup(argc,argv,flag)
int argc;
char **argv;
int flag;
{
   int found=0;
   int a,d;
   unsigned char *markoff=(unsigned char*)malloc(argc);
   int *dir=(int *)malloc(argc*sizeof(int));
   char **base=(char **)malloc(argc*sizeof(char*));
   if(!markoff || !dir || !base){
      fprintf(stderr,"%s: cannot allocate enough memory\n",progname);
      exit(1);
   }
   for(d=0;d<dir_entries;d++)matched[d]=0;
   for(a=1;a<argc;a++){
      markoff[a]=0;
      if(Mdos){
         base[a]=strrchr(argv[a],slash);
         if(base[a])base[a]++,dir[a]=finddir(argv[a],0);
         else base[a]=argv[a],dir[a]=dirno;
      }
      else base[a]=argv[a];
      if(!(base[a][0] || flag)){
         fprintf(stderr,"%s: illegal empty filename\n",progname);
         exit(1);
      }
   }
   for(d=0;d<dir_entries;d++){
      if(d==RESERVED)d+=2;
      if(!directory[d][0])continue;
      for(a=1;a<argc;a++){
         if(!base[a][0] || match(base[a],directory[d]+1)){
            if(Mdos && !ign_dirno && directory[d][254]!=dir[a])continue;
            if(!matched[d])found++,matched[d]=1;
            if(a<80)markoff[a]=1;
         }
      }
   }
   for(a=1;a<argc&&a<80;a++)
      if(!markoff[a]) fprintf(stderr,"%s not found\n",argv[a]);
   free(markoff);
   free(dir);
   free(base);
   return found;
}

int nameseq(s1,s2)
char *s1,*s2;
{
   int i;
   for(i=0;i<NAME_LEN;i++)if(!caseeq(s1[i],s2[i]))return 0;
   return 1;
}

unsigned char *makename(n)
char *n;
{
   int i;
   static unsigned char result[NAME_LEN];
   char *stripname=strrchr(n,'/');
   if(stripname)stripname++;
   else stripname=n;
   strncpy((char *)result,stripname,NAME_LEN);
   for(i=NAME_LEN;i--&&!result[i];result[i]=' ');
   return result;
}

void printname(n,f)
char *n;
FILE *f;
{
   int i,j;
   for(i=NAME_LEN;i-- && n[i]==' ';);
   if(i<0)fputs("\" \"",f);
   else for(j=0;j<=i;j++)putc(n[j],f);
}

int finddir(n,f)
char *n;
int f;           /* 1 if the last part is counted */
{
   int ans=dirno;
   static char test[NAME_LEN];
   char *tmp;
   int i,d;
   tmp=n;
   if(n[0]==slash)ans=0,n++;
   while(1){
      if(!strchr(n,slash)){
         if(!f)break;
         i=strlen(n);
         if(!i)break;
      }
      else for(i=0;i<=NAME_LEN && n[i]!=slash;i++);
      if(i==0){n++;continue;}
      if(i>NAME_LEN){
         fprintf(stderr,"%s: %s: directory name too long\n",
            progname,tmp);
         exit(1);
      }
      memcpy(test,n,i);
      n+=i;
      if(n[0])n++;
      while(i<NAME_LEN)test[i++]=' ';
      for(d=0;d<dir_entries;d++){
         if(d==RESERVED)d+=2;
         if((directory[d][0]&63)==21 && directory[d][254]==ans &&
             nameseq(test,directory[d]+1))
            break;
      }
      if(d==dir_entries){
         fprintf(stderr,"%s: %s: directory not found\n",progname,tmp);
         exit(1);
      }
      ans=directory[d][250];
   }
   return ans;
}

void findsector(f,s)
int f,s;
{
   int i=(s/800)*10+((s%800)/10)*20+(s%10);
   if(s>=1600)
      fprintf(stderr,"%s: Internal error: attempt to find sector %d\n",
              progname,s),exit(1);
   if(lseek(f,i*512,SEEK_SET)<0)samerror("lseek failed for",device);
}

void samerror(msg,name)
char *msg, *name;
{
   extern char *sys_errlist[];
   extern int sys_nerr;
   char *sysmsg;
   if(errno==EOF_MAGIC)sysmsg="End of file reached";
   else if(errno<0||errno>sys_nerr)sysmsg="Unknown error occurred";
   else sysmsg=sys_errlist[errno];
   fprintf(stderr,"%s: %s %s: %s\n",progname,msg,name,sysmsg);
   exit(1);
}
